#! /usr/bin/python3
#
#################################################################
## Copyright (c) 2013 Norbert S. <junky-zs@gmx.de>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#################################################################
#
""" Class 'cdata' for reading xml-configfile and generating dependent datastructur

read_db_config          -- reading xml-configfile and setup datastructure.
                           The 'shortname' from config-file is used as main-key 'nickname'
    # data-structur:
    #   {nickname:[{itemname:arrayindex},[value1,value2,...],Update-Flag,{itemname:logitem} ]}

configfilename          -- returns and setup of used xml-config filename.
db_sqlite_filename      -- returns and setup of used sqlite-db  filename.
db_rrdtool_filename     -- returns and setup of used rrdtool-db filename.
is_db_rrdtool_enabled   -- returns True/False for rrdtool-db config-tag 'enable'
db_rrdtool_stepseconds  -- returns value 'step_seconds' for rrdtool-db
db_rrdtool_starttime_utc-- returns value 'starttime_utc' for rrdtool-db
getlongname             -- returns the 'systempart' longname with input:'shortname'
                             Only the first two chars of 'shortname' are used, upper-
                             and lower-case is possible.
                             Undefined 'shortnames' returns: 'None'
update                  -- updates an already created data.member (logitem)
                             with the new value or will create it, if not yet available
                             parameter 'logitem' is set to 'value' (default:=0)
values                  -- returns the values (array) for 'nickname and unset 'logitem' or
                             value (single one) for 'nickname' and 'logitem'

IsAnyUpdate             -- returns True/False if any update was set
UpdateRead              -- resets status of new-update flag
IsSyspartUpdate         -- returns True/False for parameter 'nickname' if new data is
                           available and resets the flag
"""

import xml.etree.ElementTree as ET
import sys, _thread

class cdata(object):
    def __init__(self):
        self.__nickname={}
        self.__logitemHG={}
        self.__logitemHK={}
        self.__logitemWW={}
        self.__logitemSO={}
        self.__logitemDT={}
        self.__logitemNN={}
        self.__valuesHG =[]
        self.__valuesHK =[]
        self.__valuesWW =[]
        self.__valuesSO =[]
        self.__valuesDT =[]
        self.__valuesNN =[]
        self.__UpdateHG =False
        self.__UpdateHK =False
        self.__UpdateWW =False
        self.__UpdateSO =False
        self.__UpdateDT =False
        self.__UpdateNN =False
        self.__logitemMapHG={}
        self.__logitemMapHK={}
        self.__logitemMapWW={}
        self.__logitemMapSO={}
        self.__logitemMapDT={}
        self.__logitemMapNN={}
        self.__data={}
        self.__thread_lock=_thread.allocate_lock()
        self.__newdata_available=False
        self.__configfilename=""
        self.__dbname_sqlite=""
        self.__dbname_rrdtool=""
        self.__rrdtool_enable=False
        self.__rrdtool_stepseconds  =0
        self.__rrdtool_starttime_utc=0
        

    def read_db_config(self,xmlconfigpathname):
        try:
            self.__configfilename=xmlconfigpathname
            self.__tree = ET.parse(xmlconfigpathname)
        except (NameError, EnvironmentError) as e:
            print("data.read_db_config();Error;{0} on file:'{1}'".format(e.args[0], xmlconfigpathname))
            sys.exit(1)
        else:            
            try:
                self.__root = self.__tree.getroot()

                
                self.__dbname_sqlite =self.__root.find('dbname_sqlite').text
                self.__dbname_rrdtool=self.__root.find('dbname_rrd').text
                #find rrdtool-entries
                self.__dbname_rrdtool=self.__root.find('dbname_rrd').text
                for rrdtool_part in self.__root.findall('rrdtool-db'):
                    self.__rrdtool_enable       =rrdtool_part.find('enable').text.upper()
                    if self.__rrdtool_enable=='ON' or self.__rrdtool_enable=='1':
                        self.__rrdtool_enable=True
                    else:
                        self.__rrdtool_enable=False
                        
                    self.__rrdtool_stepseconds  =int(rrdtool_part.find('step_seconds').text)
                    if self.__rrdtool_stepseconds < 60:
                        self.__rrdtool_stepseconds=60
                        
                    self.__rrdtool_starttime_utc=int(rrdtool_part.find('starttime_utc').text)
                    if self.__rrdtool_starttime_utc<1344000000 or self.__rrdtool_starttime_utc>1999999999:
                        self.__rrdtool_starttime_utc=1344000000

                
                syspart=""
                for syspart in self.__root.findall('systempart'):
                    syspartname = syspart.attrib["name"]
                    shortname=""
                    for shortname in syspart.findall('shortname'):
                        shortname = shortname.attrib["name"].upper()
                    # set nicknames
                    self._setnickname(shortname, syspartname)
                    logitem=""
                    for logitem in syspart.findall('logitem'):
                        name = logitem.attrib["name"]
                        default = logitem.find('default').text
    ##### unused values from xml-file in data-context ###                    
    ##                    datatype= logitem.find('datatype').text
    ##                    datause = logitem.find('datause').text
    ##                    maxvalue= logitem.find('maxvalue').text
    ##                    unit    = logitem.find('unit').text
    ##                    values=[datatype,datause,maxvalue,default,unit]
    #####                    
                        # add itemname and values to table
                        self.update(shortname, name, default)
                    else:
                        if not len(logitem): raise AttributeError("logitem")
                else:
                    if not len(syspart): raise AttributeError("syspart")
                
            except (KeyError, IndexError, AttributeError) as e:
                if not type(e) == AttributeError:
                    print("data.read_db_config();Error on xml-tag:",e.args[0])
                else:
                    print("data.read_db_config();Error on xml-tag")                        
                sys.exit(1)
            
    def configfilename(self, xmlconfigpathname=""):
        if len(xmlconfigpathname):
            self.__configfilename=xmlconfigpathname
        return self.__configfilename
    
    def db_sqlite_filename(self, xmlconfigpathname=""):
        if len(xmlconfigpathname):
            self.__configfilename=xmlconfigpathname
            self.read_db_config(xmlconfigpathname)
        return self.__dbname_sqlite

    def db_rrdtool_filename(self, xmlconfigpathname=""):
        if len(xmlconfigpathname):
            self.__configfilename=xmlconfigpathname
            self.read_db_config(xmlconfigpathname)
        return self.__dbname_rrdtool

    def is_db_rrdtool_enabled(self):
        return self.__rrdtool_enable

    def db_rrdtool_stepseconds(self):
        return self.__rrdtool_stepseconds

    def db_rrdtool_starttime_utc(self):
        return self.__rrdtool_starttime_utc
    
    def _setnickname(self,shortname,longname):
        shortname=shortname.upper()[0:2]
        try:
            if not len(shortname): raise NameError("shortname undefined")
            if not len(longname) : raise NameError("longname undefined")
            self.__nickname.update({shortname:longname})
            if   "HG" in shortname:
                self.__data.update({"HG":[self.__logitemHG,
                                          self.__valuesHG,
                                          self.__UpdateHG,
                                          self.__logitemMapHG]})
            elif "HK" in shortname:
                self.__data.update({"HK":[self.__logitemHK,
                                          self.__valuesHK,
                                          self.__UpdateHK,
                                          self.__logitemMapHK]})
            elif "WW" in shortname:
                self.__data.update({"WW":[self.__logitemWW,
                                          self.__valuesWW,
                                          self.__UpdateWW,
                                          self.__logitemMapWW]})
            elif "SO" in shortname:
                self.__data.update({"SO":[self.__logitemSO,
                                          self.__valuesSO,
                                          self.__UpdateSO,
                                          self.__logitemMapSO]})
            elif "DT" in shortname:
                self.__data.update({"DT":[self.__logitemDT,
                                          self.__valuesDT,
                                          self.__UpdateDT,
                                          self.__logitemMapDT]})
            else:
                self.__data.update({"NN":[self.__logitemNN,
                                          self.__valuesNN,
                                          self.__UpdateNN,
                                          self.__logitemMapNN]})
            
        except (NameError, AttributeError) as e:
            print("data._setnickname();Error;{0}".format(e.args[0]))

            
    def getlongname(self,shortname):
        try:
            return self.__nickname[shortname.upper()[0:2]]
        except (KeyError, IndexError, AttributeError) as e:
            print("data.getlongname();Error;longname'{0}' not found".format(e.args[0]))
            return None

    ############################
    # function updates tableentry, if not yet available tableentry is created and set
    #
    # data-structur:
    #   {nickname:[ {itemname:arrayindex},[value1,value2,...],Update-Flag,{itemname:logitem} ]}
    #
    def update(self, nickname, logitem, value=0.0):
        nickname=nickname.upper()[0:2]
        itemname=logitem.replace("_","").lower()
        try:
            if not len(nickname): raise NameError("nickname undefined")
            if not len(itemname): raise NameError("itemname undefined")
            if nickname in self.__nickname:
                if itemname in self.__data[nickname][0]:
                    # update value, get index first from dir{itemname:index}
                    index=int(self.__data[nickname][0][itemname])
                    self.__data[nickname][1][index]=value
                    # set 'IsSyspartUpdate' true
                    self.__data[nickname][2]=True
                else:
                    # add new item and value, index is current array-length
                    # and is set to dir{itemname:index}
                    index=len(self.__data[nickname][1])
                    self.__data[nickname][0].update({itemname:index})
                    self.__data[nickname][1].append(value)
                    # set internal itemname to external logitem
                    self.__data[nickname][3].update({itemname:logitem})
            else:
                raise NameError("nickname:'{0}' not found".format(nickname))
            self.__thread_lock.acquire()
            self.__newdata_available=True
            self.__thread_lock.release()
        except (KeyError, NameError, AttributeError) as e:
            print("data.update();Error;{0}".format(e.args[0]))
            self.__thread_lock.acquire()
            self.__newdata_available=False
            self.__thread_lock.release()
                
    def values(self, nickname, logitem=""):
        nickname=nickname.upper()[0:2]
        itemname=logitem.replace("_","").lower()
        try:
            self.__thread_lock.acquire()
            if not len(nickname): raise NameError("nickname undefined")
            if nickname in self.__nickname:
                if len(itemname):
                    if itemname in self.__data[nickname][0]:
                        
                        index=int(self.__data[nickname][0][itemname])
                        return self.__data[nickname][1][index]
                    else:
                        raise NameError("itemname:'{0}' not found".format(nickname))
                else:
                    return self.__data[nickname][1]
            else:
                raise NameError("nickname:'{0}' not found".format(nickname))
        except(KeyError, NameError, AttributeError) as e:
            print("data.values();Error;{0}".format(e.args[0]))

        finally:
            self.__thread_lock.release()

    def IsAnyUpdate(self):
        return self.__newdata_available

    def UpdateRead(self):
        self.__thread_lock.acquire()
        self.__newdata_available=False
        self.__thread_lock.release()

    def IsSyspartUpdate(self, nickname):
        nickname=nickname.upper()[0:2]
        Rtn=self.__data[nickname][2]
        self.__data[nickname][2]=False
        return Rtn

#--- class cdata end ---#
        
if __name__ == "__main__":
    ########
    ##
    # testtype:= 0 -> basic 'cdata'-classtest
    # testtype:= 1 -> test including xml-configfile-data extract
    #
    # data-structur:
    #   {nickname:[ {itemname:arrayindex},[value1,value2,...],Update-Flag,{itemname:logitem} ]}
    #
    testtype=1
    
    if testtype == 0:
        #---------------- testtype=0 -------------------
        data=cdata()
        print("-- setup nicknames --")
        data._setnickname("HG","heizgeraet")
        data._setnickname("Hklein","heizkreis")
        data._setnickname("WW","warmwasser")
        data._setnickname("SO","solar")
        data._setnickname("DT","datetime")
        data._setnickname("data","dadadata")
        #----------------
        print("-- updata data --")
        data.update("SO","T_Collector")
        data.update("SO","T_Collector",33.5)
        data.update("SO","T_Soll")
        data.update("SO","T_Soll",13.4)
        data.update("SO","T_Speicherunten",20.2)
        data.update("SO","T_Laufzeit",3456)
        data.update("SO","T_Pumpe",1)
        data.update("HK","T_Ist",43.8)
        data.update("SO","T_Aussen",0.2)
        data.update("DT","Date","23.03.2013")
        data.update("DT","Time","11:12:13")
        print("----------- unknown values check ----------")
        print(" -- setup nickname with wrong value --")
        data._setnickname("","solar")
        data._setnickname("WW","")
        print(" -- check unavailable names    --")
        data.update("ST","T_Collector")
        data.update("","T_Collector")
        data.update("ST","")
        print(' -- check for unknown longname --')
        print(data.getlongname("LO"))
        print(""" -- check for unknown values in ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO","T_blabla"))
        print(data.values("","T_Pumpe"))
        #----------------
        print("----------- normal operation   ----------")
        print("-- Known values get and update --")

        print(""" -- get: all values from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO"))
        print(""" -- get: value:'T_Laufzeit' from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO","T_Laufzeit"))
        print(""" -- update: value:'T_Laufzeit' at ["{0}"] to 654321 ---""".format(
                                    data.getlongname("SO")))
        data.update("SO","T_Laufzeit",654321)
        print(""" -- again get: value:'T_Laufzeit' from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO","T_Laufzeit"))
        print(""" -- get: values:'Date' and 'Time' from ["{0}"] ---""".format(
                                    data.getlongname("DT")))
        print("{0} {1}".format(data.values("DT","Date"), data.values("DT","Time")))
    else:
        #---------------- testtype=1 -------------------
        import os
        print("current path   :'{0}'".format(os.getcwd()))

        data=cdata()
        data.read_db_config("./../etc/config/4test/create_db_test.xml")

        print("configfile            :'{0}'".format(data.configfilename()))
        print("Sqlite  db-file       :'{0}'".format(data.db_sqlite_filename()))
        print("rrdtool db-file       :'{0}'".format(data.db_rrdtool_filename()))
        print("rrdtool db_enabled    :{0}".format(data.is_db_rrdtool_enabled()))
        print("rrdtool db_stepseconds:{0}".format(data.db_rrdtool_stepseconds()))
        print("rrdtool db_starttime  :{0}".format(data.db_rrdtool_starttime_utc()))
        
        
        print(""" -- get: all values from ["{0}"] ---""".format(
                                    data.getlongname("HG")))
        print(data.values("HG"))
        print(""" -- get: all values from ["{0}"] ---""".format(
                                    data.getlongname("HK")))
        print(data.values("HK"))
        print(""" -- get: all values from ["{0}"] ---""".format(
                                    data.getlongname("WW")))
        print(data.values("WW"))
        print(""" -- get: all values from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO"))

        print(""" -- get       value:'T_Kollektor' from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO","T_Kollektor"))
        print(""" -- update    value:'T_Kollektor' at   ["{0}"] to 89.1 ---""".format(
                                    data.getlongname("SO")))
        data.update("SO","T_Kollektor",89.1)
        print(""" -- again get value:'T_Kollektor' from ["{0}"] ---""".format(
                                    data.getlongname("SO")))
        print(data.values("SO","T_Kollektor"))
